// Copyright 2024 PingCAP, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package injectcall

import (
	"context"
	"fmt"

	"github.com/pingcap/failpoint"
)

func foo(ctx context.Context, count int) int {
	for i := 0; i < count; i++ {
		fmt.Println(i)
		failpoint.InjectCall("test", ctx, i, count)
		select {
		case <-ctx.Done():
			return i
		default:
		}
	}
	return count
}
