// Copyright 2019 PingCAP, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package version

import "fmt"

var (
	releaseVersion = "None"
	buildTS        = "None"
	gitHash        = "None"
	gitBranch      = "None"
	goVersion      = "None"
)

// PrintVersion prints the release information
func PrintVersion() {
	fmt.Println("ReleaseVersion", releaseVersion)
	fmt.Println("BuildTS", buildTS)
	fmt.Println("GitHash", gitHash)
	fmt.Println("GitBranch", gitBranch)
	fmt.Println("GoVersion", goVersion)
}
